/*
 * arch/arm/mach-ep93xx/clock.c
 * Clock control for Cirrus EP93xx chips.
 *
 * Copyright (C) 2006 Lennert Buytenhek <buytenh@wantstofly.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 */

#include <linux/kernel.h>
#include <linux/clk.h>
#include <linux/err.h>
#include <linux/module.h>
#include <linux/string.h>
#include <linux/io.h>

#include <asm/clkdev.h>
#include <asm/div64.h>
#include <mach/hardware.h>


/*
 * The EP93xx has two external crystal oscillators.  To generate the
 * required high-frequency clocks, the processor uses two phase-locked-
 * loops (PLLs) to multiply the incoming external clock signal to much
 * higher frequencies that are then divided down by programmable dividers
 * to produce the needed clocks.  The PLLs operate independently of one
 * another.
 */
#define EP93XX_EXT_CLK_RATE	14745600
#define EP93XX_EXT_RTC_RATE	32768

struct clk {
	unsigned long	rate;
	int		users;
	u32		enable_reg;
	u32		enable_mask;
	int		write_locked;

	int		(*set_rate)(struct clk *clk, unsigned long rate);
};


static int set_ssp_rate(struct clk *clk, unsigned long rate);
static int set_div_rate(struct clk *clk, unsigned long rate);

static struct clk clk_uart = {
	.rate		= EP93XX_EXT_CLK_RATE,
};
static struct clk clk_pll1;
static struct clk clk_f;
static struct clk clk_h;
static struct clk clk_p;
static struct clk clk_pll2;
static struct clk clk_usb_host = {
	.enable_reg	= EP93XX_SYSCON_CLOCK_CONTROL,
	.enable_mask	= EP93XX_SYSCON_CLOCK_USH_EN,
};
static struct clk clk_ssp;
static struct clk ssp_clkout = {
	.set_rate	= set_ssp_rate,
};

static struct clk clk_vid = {
	.enable_reg     = EP93XX_SYSCON_VIDCLKDIV,
	.enable_mask    = EP93XX_SYSCON_CLKDIV_ENABLE,
	.write_locked   = 1,
	.set_rate	= set_div_rate,
};

static struct clk clk_i2s = {
	.enable_reg	= EP93XX_SYSCON_I2SCLKDIV,
	.enable_mask	= EP93XX_SYSCON_CLKDIV_ENABLE,
	.write_locked	= 1,
	.set_rate	= set_div_rate,
};

/* DMA Clocks */
static struct clk clk_m2p0 = {
	.enable_reg	= EP93XX_SYSCON_CLOCK_CONTROL,
	.enable_mask	= 0x00020000,
};
static struct clk clk_m2p1 = {
	.enable_reg	= EP93XX_SYSCON_CLOCK_CONTROL,
	.enable_mask	= 0x00010000,
};
static struct clk clk_m2p2 = {
	.enable_reg	= EP93XX_SYSCON_CLOCK_CONTROL,
	.enable_mask	= 0x00080000,
};
static struct clk clk_m2p3 = {
	.enable_reg	= EP93XX_SYSCON_CLOCK_CONTROL,
	.enable_mask	= 0x00040000,
};
static struct clk clk_m2p4 = {
	.enable_reg	= EP93XX_SYSCON_CLOCK_CONTROL,
	.enable_mask	= 0x00200000,
};
static struct clk clk_m2p5 = {
	.enable_reg	= EP93XX_SYSCON_CLOCK_CONTROL,
	.enable_mask	= 0x00100000,
};
static struct clk clk_m2p6 = {
	.enable_reg	= EP93XX_SYSCON_CLOCK_CONTROL,
	.enable_mask	= 0x00800000,
};
static struct clk clk_m2p7 = {
	.enable_reg	= EP93XX_SYSCON_CLOCK_CONTROL,
	.enable_mask	= 0x00400000,
};
static struct clk clk_m2p8 = {
	.enable_reg	= EP93XX_SYSCON_CLOCK_CONTROL,
	.enable_mask	= 0x02000000,
};
static struct clk clk_m2p9 = {
	.enable_reg	= EP93XX_SYSCON_CLOCK_CONTROL,
	.enable_mask	= 0x01000000,
};
static struct clk clk_m2m0 = {
	.enable_reg	= EP93XX_SYSCON_CLOCK_CONTROL,
	.enable_mask	= 0x04000000,
};
static struct clk clk_m2m1 = {
	.enable_reg	= EP93XX_SYSCON_CLOCK_CONTROL,
	.enable_mask	= 0x08000000,
};

#define INIT_CK(dev,con,ck)					\
	{ .dev_id = dev, .con_id = con, .clk = ck }

static struct clk_lookup clocks[] = {
	INIT_CK("apb:uart1",	NULL,		&clk_uart),
	INIT_CK("apb:uart2",	NULL,		&clk_uart),
	INIT_CK("apb:uart3",	NULL,		&clk_uart),
	INIT_CK(NULL,		"pll1",		&clk_pll1),
	INIT_CK(NULL,		"fclk",		&clk_f),
	INIT_CK(NULL,		"hclk",		&clk_h),
	INIT_CK(NULL,		"pclk",		&clk_p),
	INIT_CK(NULL,		"pll2",		&clk_pll2),
	INIT_CK(NULL,		"usb_host",	&clk_usb_host),
	INIT_CK(NULL,		"ssp_clk",	&clk_ssp),
	INIT_CK("ep93xx-spi.0",	NULL,		&ssp_clkout),
	INIT_CK("ep93xxfb.0",	NULL,		&clk_vid),
	INIT_CK(NULL,		"i2s",		&clk_i2s),
	INIT_CK(NULL,		"m2p0",		&clk_m2p0),
	INIT_CK(NULL,		"m2p1",		&clk_m2p1),
	INIT_CK(NULL,		"m2p2",		&clk_m2p2),
	INIT_CK(NULL,		"m2p3",		&clk_m2p3),
	INIT_CK(NULL,		"m2p4",		&clk_m2p4),
	INIT_CK(NULL,		"m2p5",		&clk_m2p5),
	INIT_CK(NULL,		"m2p6",		&clk_m2p6),
	INIT_CK(NULL,		"m2p7",		&clk_m2p7),
	INIT_CK(NULL,		"m2p8",		&clk_m2p8),
	INIT_CK(NULL,		"m2p9",		&clk_m2p9),
	INIT_CK(NULL,		"m2m0",		&clk_m2m0),
	INIT_CK(NULL,		"m2m1",		&clk_m2m1),
};


int clk_enable(struct clk *clk)
{
	if (!clk->users++ && clk->enable_reg) {
		u32 value;

		value = __raw_readl(clk->enable_reg) | clk->enable_mask;
		if (clk->write_locked)
			ep93xx_syscon_swlocked_write(clk->enable_reg, value);
		else
			__raw_writel(value, clk->enable_reg);
	}

	return 0;
}
EXPORT_SYMBOL(clk_enable);

void clk_disable(struct clk *clk)
{
	if (!--clk->users && clk->enable_reg) {
		u32 value;
		
		value = __raw_readl(clk->enable_reg) & ~clk->enable_mask;
		if (clk->write_locked)
			ep93xx_syscon_swlocked_write(clk->enable_reg, value);
		else
			__raw_writel(value, clk->enable_reg);
	}
}
EXPORT_SYMBOL(clk_disable);

unsigned long clk_get_rate(struct clk *clk)
{
	return clk->rate;
}
EXPORT_SYMBOL(clk_get_rate);

static int set_ssp_rate(struct clk *clk, unsigned long rate)
{
	unsigned long sspclk = clk_ssp.rate;
	unsigned long sspclkout;
	unsigned int cpsdvr, scr;
	u32 value;

	/* Calculate the closest attainable rate */
	for (cpsdvr = 2; cpsdvr < 255; cpsdvr++) {
		for (scr = 0; scr < 256; scr++) {
			sspclkout = sspclk / (cpsdvr * (1 + scr));
			if (sspclkout <= rate)
				break;
		}
		if (sspclkout <= rate)
			break;
	}

	/* We only need to set the clock if it's changed */
	if (sspclkout != clk->rate) {
		__raw_writel(cpsdvr, EP93XX_SPI_SSPCPSR);
		value = __raw_readl(EP93XX_SPI_SSPCR0) & 0xff;
		value |= (scr << 8);
		__raw_writel(value, EP93XX_SPI_SSPCR0);

		clk->rate = sspclkout;
	}

	return 0;
}

static unsigned long calc_clk_div(unsigned long rate, int *psel, int *esel, 
				  int *pdiv, int *div)
{
	unsigned long max_rate, set_rate = 0, actual_rate = 0, 
		mclk_rate = 0, rate_err = -1;
	int i, found = 0, __div = 0, __pdiv = 0;
	
	/* Don't exceed the maximum rate */
	max_rate = max(max(clk_pll1.rate / 4, clk_pll2.rate / 4),
		       (unsigned long)EP93XX_EXT_CLK_RATE / 4);
	rate = min(rate, max_rate);
	
	/*
	 * Try the two pll's and the external clock
	 * Because the valid predividers are 2, 2.5 and 3, we multiply
	 * all the clocks by 2 to avoid floating point math.
	 *
	 * This is based on the algorithm in the ep93xx raster guide:
	 * http://be-a-maverick.com/en/pubs/appNote/AN269REV1.pdf
	 *
	 */
	for (i = 0; i < 3; i++) {
		if (i == 0)
			mclk_rate = EP93XX_EXT_CLK_RATE * 2;
		else if (i == 1)
			mclk_rate = clk_pll1.rate * 2;
		else if (i == 2)
			mclk_rate = clk_pll2.rate * 2;
		
		/* Try each predivider value */
		for (__pdiv = 4; __pdiv <= 6; __pdiv++) {
			__div = mclk_rate / (rate * __pdiv);
			if (__div < 2 || __div > 127)
				continue;
			
			actual_rate = mclk_rate / (__pdiv * __div);
			pr_debug("rate = %lu, mclk = %lu, pdiv = %d, "
				 "div = %d, src = %s, actual = %lu\n",
				 rate, mclk_rate, __pdiv, __div, 
				 i == 0 ? "XTALI" : i == 1 ? "PLL1" : "PLL2",
				 actual_rate);
			
			if (!found || abs(actual_rate - rate) < rate_err) {
				*pdiv = __pdiv - 3;
				*div = __div;
				*psel = (i == 2);
				*esel = (i != 0);
				set_rate = actual_rate;
				rate_err = actual_rate - rate;
				found = 1;
			}
		}
	}
	
	if (!found)
		return 0;
	
	return set_rate;
}

static int set_div_rate(struct clk *clk, unsigned long rate)
{
	unsigned long actual_rate;
	int psel = 0, esel = 0, pdiv = 0, div = 0;
	u32 val;

	actual_rate = calc_clk_div(rate, &psel, &esel, &pdiv, &div);
	if (actual_rate == 0)
		return -EINVAL;
	clk->rate = actual_rate;

	/* Clear the esel, psel, pdiv and div bits */
	val = __raw_readl(clk->enable_reg);
	val &= ~0x637f;
	ep93xx_syscon_swlocked_write(clk->enable_reg, val);
	
	/* Set the new esel, psel, pdiv and div bits for the new clock rate */
	val |= (esel ? EP93XX_SYSCON_CLKDIV_ESEL : 0) |
		(psel ? EP93XX_SYSCON_CLKDIV_PSEL : 0) |
		(pdiv << EP93XX_SYSCON_CLKDIV_PDIV_SHIFT) | div;
	ep93xx_syscon_swlocked_write(clk->enable_reg, val);

	return 0;
}

int clk_set_rate(struct clk *clk, unsigned long rate)
{
	if (clk->set_rate)
		return clk->set_rate(clk, rate);

	return -EINVAL;
}
EXPORT_SYMBOL(clk_set_rate);


static char fclk_divisors[] = { 1, 2, 4, 8, 16, 1, 1, 1 };
static char hclk_divisors[] = { 1, 2, 4, 5, 6, 8, 16, 32 };
static char pclk_divisors[] = { 1, 2, 4, 8 };
static char *ep93xx_rev[] = { "A", "B", "C", "D0", "D1", "E0", "E1", "E2" };

/*
 * PLL rate = 14.7456 MHz * (X1FBD + 1) * (X2FBD + 1) / (X2IPD + 1) / 2^PS
 */
static unsigned long calc_pll_rate(u32 config_word)
{
	unsigned long long rate;
	int i;

	rate = EP93XX_EXT_CLK_RATE;
	rate *= ((config_word >> 11) & 0x1f) + 1;		/* X1FBD */
	rate *= ((config_word >> 5) & 0x3f) + 1;		/* X2FBD */
	do_div(rate, (config_word & 0x1f) + 1);			/* X2IPD */
	for (i = 0; i < ((config_word >> 16) & 3); i++)		/* PS */
		rate >>= 1;

	return (unsigned long)rate;
}

static void __init ep93xx_dma_clock_init(void)
{
	clk_m2p0.rate = clk_h.rate;
	clk_m2p1.rate = clk_h.rate;
	clk_m2p2.rate = clk_h.rate;
	clk_m2p3.rate = clk_h.rate;
	clk_m2p4.rate = clk_h.rate;
	clk_m2p5.rate = clk_h.rate;
	clk_m2p6.rate = clk_h.rate;
	clk_m2p7.rate = clk_h.rate;
	clk_m2p8.rate = clk_h.rate;
	clk_m2p9.rate = clk_h.rate;
	clk_m2m0.rate = clk_h.rate;
	clk_m2m1.rate = clk_h.rate;
}

static int __init ep93xx_clock_init(void)
{
	u32 value;
	int i;

	value = __raw_readl(EP93XX_SYSCON_CLOCK_SET1);
	if (!(value & 0x00800000)) {			/* PLL1 bypassed?  */
		clk_pll1.rate = EP93XX_EXT_CLK_RATE;
	} else {
		clk_pll1.rate = calc_pll_rate(value);
	}
	clk_f.rate = clk_pll1.rate / fclk_divisors[(value >> 25) & 0x7];
	clk_h.rate = clk_pll1.rate / hclk_divisors[(value >> 20) & 0x7];
	clk_p.rate = clk_h.rate / pclk_divisors[(value >> 18) & 0x3];
	ep93xx_dma_clock_init();

	value = __raw_readl(EP93XX_SYSCON_CLOCK_SET2);
	if (!(value & 0x00080000)) {			/* PLL2 bypassed?  */
		clk_pll2.rate = EP93XX_EXT_CLK_RATE;
	} else if (value & 0x00040000) {		/* PLL2 enabled?  */
		clk_pll2.rate = calc_pll_rate(value);
	} else {
		clk_pll2.rate = 0;
	}
	clk_usb_host.rate = clk_pll2.rate / (((value >> 28) & 0xf) + 1);

	/* See http://www.cirrus.com/en/pubs/appNote/AN273REV4.pdf */
	value = __raw_readl(EP93XX_SYSCON_CHIP_ID);
	value = (value >> 28) & 0xf;
	if (value < 7)
		clk_ssp.rate = EP93XX_EXT_CLK_RATE / 2;
	else
		clk_ssp.rate = EP93XX_EXT_CLK_RATE;

	printk(KERN_INFO "ep93xx: PLL1 running at %ld MHz, PLL2 at %ld MHz\n",
		clk_pll1.rate / 1000000, clk_pll2.rate / 1000000);
	printk(KERN_INFO "ep93xx: FCLK %ld MHz, HCLK %ld MHz, PCLK %ld MHz\n",
		clk_f.rate / 1000000, clk_h.rate / 1000000,
		clk_p.rate / 1000000);
	if (value < 8)
		printk(KERN_INFO "ep93xx: Silicon Rev %s\n", ep93xx_rev[value]);
	else
		printk(KERN_INFO "ep93xx: Silicon Rev unknown (%d)\n", value);

	for (i = 0; i < ARRAY_SIZE(clocks); i++)
		clkdev_add(&clocks[i]);
	return 0;
}
arch_initcall(ep93xx_clock_init);
