/*
 *  drivers/net/mv88e6061.c
 *
 *  Author:     Theuns Verwoerd <theuns@bluewatersys.com>
 *  Created:    25 Feb 2009
 *  Copyright:  Bluewater Systems Ltd
 *
 *	Marvell 88E6061 Low Power 6-port 10/100 Switch MDIO driver
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
 
#include <linux/device.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <asm/io.h>
 
#define DRV_MODULE_NAME		"mv88e6061-switch"
#define DRV_MODULE_VERSION	"0.1"

// TODO hang this off a platform device
static struct mv88e6061_priv
{
	void			*base_addr;
} mv88e6061_priv;

static struct mv88e6061_priv *mv88e6061_ptr = &mv88e6061_priv;

#define REG_MIICMD		0x0010
#define  REG_MIICMD_READ	0x00008000
#define  REG_MIICMD_WRITE	0x00004000
#define REG_MIIDATA		0x0014
#define REG_MIISTS		0x0018
#define  REG_MIISTS_BUSY	0x00000001

#define rdb(ep, off)		__raw_readb((ep)->base_addr + (off))
#define rdw(ep, off)		__raw_readw((ep)->base_addr + (off))
#define rdl(ep, off)		__raw_readl((ep)->base_addr + (off))
#define wrb(ep, off, val)	__raw_writeb((val), (ep)->base_addr + (off))
#define wrw(ep, off, val)	__raw_writew((val), (ep)->base_addr + (off))
#define wrl(ep, off, val)	__raw_writel((val), (ep)->base_addr + (off))

//static int mdio_read(struct net_device *dev, int phy_id, int reg)
static int mdio_read(void *dev, int phy_id, int reg)
{
//	struct ep93xx_priv *ep = netdev_priv(dev);
	int data = 0, ret;
	int i;

//	printk ("\n%s:%u dev %p id %x reg %x", __FUNCTION__, __LINE__, dev, phy_id, reg);
	
	phy_id &= 0x1f;
	reg &= 0x1f;
	
	wrl(mv88e6061_ptr, REG_MIICMD, REG_MIICMD_READ | (phy_id << 5) | reg);
	
//	printk (" wrote %08x", REG_MIICMD_READ | (phy_id << 5) | reg);

	for (i = 0; i < 10; i++) {
		if ((rdl(mv88e6061_ptr, REG_MIISTS) & REG_MIISTS_BUSY) == 0)
			break;
		msleep(1);
	}

	if (i == 10) {
		printk(KERN_INFO DRV_MODULE_NAME ": mdio read timed out\n");
		data = 0xffff;
	} else do {
		/* We seem to have some problems reading, so keep trying until it works */
		ret = data;
		data = rdl(mv88e6061_ptr, REG_MIIDATA);
		i++;
	} while ((data == 0xFFF) && (i < 20)); //(data != ret);

//	printk(" ret %x\n", data);
	
	return data;
}

//static void mdio_write(struct net_device *dev, int phy_id, int reg, int data)
static void mdio_write(void *dev, int phy_id, int reg, int data)
{
//	struct ep93xx_priv *ep = netdev_priv(dev);
	int i;

//	printk ("\n%s:%u dev %p id %x reg %x", __FUNCTION__, __LINE__, dev, phy_id, reg);
	
	wrl(mv88e6061_ptr, REG_MIIDATA, data);
	wrl(mv88e6061_ptr, REG_MIICMD, REG_MIICMD_WRITE | (phy_id << 5) | reg);

	for (i = 0; i < 10; i++) {
		if ((rdl(mv88e6061_ptr, REG_MIISTS) & REG_MIISTS_BUSY) == 0)
			break;
		msleep(1);
	}

	/* Write twice to ensure latching doesn't interfere */
	wrl(mv88e6061_ptr, REG_MIIDATA, data);
	wrl(mv88e6061_ptr, REG_MIICMD, REG_MIICMD_WRITE | (phy_id << 5) | reg);

	for (i = 0; i < 10; i++) {
		if ((rdl(mv88e6061_ptr, REG_MIISTS) & REG_MIISTS_BUSY) == 0)
			break;
		msleep(1);
	}
	
//	printk(" wrote %x\n", data);
	
	if (i == 10)
		printk(KERN_INFO DRV_MODULE_NAME ": mdio write timed out\n");
}


static ssize_t show_raw_real(struct device *dev, struct device_attribute *attr, char *buf, int phy, int reg)
{
	u32	regVal;
	
	// TODO locking
	regVal = mdio_read(dev, phy, reg);
	return sprintf(buf, "0x%08x\n", regVal);
}

#define show_raw(name, phy, reg)	\
static ssize_t show_raw_##name(struct device *dev, struct device_attribute *attr, char *buf)	\
{																								\
	return show_raw_real(dev, attr, buf, phy, reg);												\
}

static ssize_t set_raw_real(struct device *dev, struct device_attribute *attr, const char *buf, size_t count, int phy, int reg)
{
	int regVal = simple_strtoul(buf, NULL, 0);

	mdio_write(dev, phy, reg, regVal);
	return count;
}

#define set_raw(name, phy, reg)	\
static ssize_t set_raw_##name(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)	\
{																		\
	return set_raw_real(dev, attr, buf, count, phy, reg);					\
}

#define set_show_etc(name, phy, reg, w) 	\
	show_raw(name, phy, reg);				\
	set_raw(name, phy, reg);				\
	static DEVICE_ATTR(name, S_IRUGO | (w ? S_IWUSR : 0), show_raw_##name, set_raw_##name);

/* Per-port registers */
set_show_etc(0_port_status,		0x8 /*phy*/, 0 /*reg*/, 1);
set_show_etc(0_mac_control,		0x8 /*phy*/, 1 /*reg*/, 1);
set_show_etc(0_switch_id,		0x8 /*phy*/, 3 /*reg*/, 1);
set_show_etc(0_port_control,	0x8 /*phy*/, 4 /*reg*/, 1);
set_show_etc(0_port_vlan_map,	0x8 /*phy*/, 6 /*reg*/, 1);
set_show_etc(0_def_vlan_ID_pri,	0x8 /*phy*/, 7 /*reg*/, 1);
set_show_etc(0_port_control2,	0x8 /*phy*/, 8 /*reg*/, 1);
set_show_etc(0_ingr_rate_ctrl,	0x8 /*phy*/, 9 /*reg*/, 1);
set_show_etc(0_egr_rate_ctrl,	0x8 /*phy*/,10 /*reg*/, 1);
set_show_etc(0_port_assoc_vec,	0x8 /*phy*/,11 /*reg*/, 1);
set_show_etc(0_rx_frame_cnt,	0x8 /*phy*/,16 /*reg*/, 1);
set_show_etc(0_tx_frame_cnt,	0x8 /*phy*/,17 /*reg*/, 1);
set_show_etc(0_rx_pol_cnt,		0x8 /*phy*/,18 /*reg*/, 1);
set_show_etc(0_queue_cnt,		0x8 /*phy*/,27 /*reg*/, 1);

set_show_etc(1_port_status,		0x9 /*phy*/, 0 /*reg*/, 1);
set_show_etc(1_mac_control,		0x9 /*phy*/, 1 /*reg*/, 1);
set_show_etc(1_switch_id,		0x9 /*phy*/, 3 /*reg*/, 1);
set_show_etc(1_port_control,	0x9 /*phy*/, 4 /*reg*/, 1);
set_show_etc(1_port_vlan_map,	0x9 /*phy*/, 6 /*reg*/, 1);
set_show_etc(1_def_vlan_ID_pri,	0x9 /*phy*/, 7 /*reg*/, 1);
set_show_etc(1_port_control2,	0x9 /*phy*/, 8 /*reg*/, 1);
set_show_etc(1_ingr_rate_ctrl,	0x9 /*phy*/, 9 /*reg*/, 1);
set_show_etc(1_egr_rate_ctrl,	0x9 /*phy*/,10 /*reg*/, 1);
set_show_etc(1_port_assoc_vec,	0x9 /*phy*/,11 /*reg*/, 1);
set_show_etc(1_rx_frame_cnt,	0x9 /*phy*/,16 /*reg*/, 1);
set_show_etc(1_tx_frame_cnt,	0x9 /*phy*/,17 /*reg*/, 1);
set_show_etc(1_rx_pol_cnt,		0x9 /*phy*/,18 /*reg*/, 1);
set_show_etc(1_queue_cnt,		0x9 /*phy*/,27 /*reg*/, 1);

set_show_etc(2_port_status,		0xa /*phy*/, 0 /*reg*/, 1);
set_show_etc(2_mac_control,		0xa /*phy*/, 1 /*reg*/, 1);
set_show_etc(2_switch_id,		0xa /*phy*/, 3 /*reg*/, 1);
set_show_etc(2_port_control,	0xa /*phy*/, 4 /*reg*/, 1);
set_show_etc(2_port_vlan_map,	0xa /*phy*/, 6 /*reg*/, 1);
set_show_etc(2_def_vlan_ID_pri,	0xa /*phy*/, 7 /*reg*/, 1);
set_show_etc(2_port_control2,	0xa /*phy*/, 8 /*reg*/, 1);
set_show_etc(2_ingr_rate_ctrl,	0xa /*phy*/, 9 /*reg*/, 1);
set_show_etc(2_egr_rate_ctrl,	0xa /*phy*/,10 /*reg*/, 1);
set_show_etc(2_port_assoc_vec,	0xa /*phy*/,11 /*reg*/, 1);
set_show_etc(2_rx_frame_cnt,	0xa /*phy*/,16 /*reg*/, 1);
set_show_etc(2_tx_frame_cnt,	0xa /*phy*/,17 /*reg*/, 1);
set_show_etc(2_rx_pol_cnt,		0xa /*phy*/,18 /*reg*/, 1);
set_show_etc(2_queue_cnt,		0xa /*phy*/,27 /*reg*/, 1);

set_show_etc(3_port_status,		0xb /*phy*/, 0 /*reg*/, 1);
set_show_etc(3_mac_control,		0xb /*phy*/, 1 /*reg*/, 1);
set_show_etc(3_switch_id,		0xb /*phy*/, 3 /*reg*/, 1);
set_show_etc(3_port_control,	0xb /*phy*/, 4 /*reg*/, 1);
set_show_etc(3_port_vlan_map,	0xb /*phy*/, 6 /*reg*/, 1);
set_show_etc(3_def_vlan_ID_pri,	0xb /*phy*/, 7 /*reg*/, 1);
set_show_etc(3_port_control2,	0xb /*phy*/, 8 /*reg*/, 1);
set_show_etc(3_ingr_rate_ctrl,	0xb /*phy*/, 9 /*reg*/, 1);
set_show_etc(3_egr_rate_ctrl,	0xb /*phy*/,10 /*reg*/, 1);
set_show_etc(3_port_assoc_vec,	0xb /*phy*/,11 /*reg*/, 1);
set_show_etc(3_rx_frame_cnt,	0xb /*phy*/,16 /*reg*/, 1);
set_show_etc(3_tx_frame_cnt,	0xb /*phy*/,17 /*reg*/, 1);
set_show_etc(3_rx_pol_cnt,		0xb /*phy*/,18 /*reg*/, 1);
set_show_etc(3_queue_cnt,		0xb /*phy*/,27 /*reg*/, 1);

set_show_etc(4_port_status,		0xc /*phy*/, 0 /*reg*/, 1);
set_show_etc(4_mac_control,		0xc /*phy*/, 1 /*reg*/, 1);
set_show_etc(4_switch_id,		0xc /*phy*/, 3 /*reg*/, 1);
set_show_etc(4_port_control,	0xc /*phy*/, 4 /*reg*/, 1);
set_show_etc(4_port_vlan_map,	0xc /*phy*/, 6 /*reg*/, 1);
set_show_etc(4_def_vlan_ID_pri,	0xc /*phy*/, 7 /*reg*/, 1);
set_show_etc(4_port_control2,	0xc /*phy*/, 8 /*reg*/, 1);
set_show_etc(4_ingr_rate_ctrl,	0xc /*phy*/, 9 /*reg*/, 1);
set_show_etc(4_egr_rate_ctrl,	0xc /*phy*/,10 /*reg*/, 1);
set_show_etc(4_port_assoc_vec,	0xc /*phy*/,11 /*reg*/, 1);
set_show_etc(4_rx_frame_cnt,	0xc /*phy*/,16 /*reg*/, 1);
set_show_etc(4_tx_frame_cnt,	0xc /*phy*/,17 /*reg*/, 1);
set_show_etc(4_rx_pol_cnt,		0xc /*phy*/,18 /*reg*/, 1);
set_show_etc(4_queue_cnt,		0xc /*phy*/,27 /*reg*/, 1);

set_show_etc(5_port_status,		0xd /*phy*/, 0 /*reg*/, 1);
set_show_etc(5_mac_control,		0xd /*phy*/, 1 /*reg*/, 1);
set_show_etc(5_switch_id,		0xd /*phy*/, 3 /*reg*/, 1);
set_show_etc(5_port_control,	0xd /*phy*/, 4 /*reg*/, 1);
set_show_etc(5_port_vlan_map,	0xd /*phy*/, 6 /*reg*/, 1);
set_show_etc(5_def_vlan_ID_pri,	0xd /*phy*/, 7 /*reg*/, 1);
set_show_etc(5_port_control2,	0xd /*phy*/, 8 /*reg*/, 1);
set_show_etc(5_ingr_rate_ctrl,	0xd /*phy*/, 9 /*reg*/, 1);
set_show_etc(5_egr_rate_ctrl,	0xd /*phy*/,10 /*reg*/, 1);
set_show_etc(5_port_assoc_vec,	0xd /*phy*/,11 /*reg*/, 1);
set_show_etc(5_rx_frame_cnt,	0xd /*phy*/,16 /*reg*/, 1);
set_show_etc(5_tx_frame_cnt,	0xd /*phy*/,17 /*reg*/, 1);
set_show_etc(5_rx_pol_cnt,		0xd /*phy*/,18 /*reg*/, 1);
set_show_etc(5_queue_cnt,		0xd /*phy*/,27 /*reg*/, 1);

/* Global registers */
set_show_etc(trunk_mask,		0xe /*phy*/, 7 /*reg*/, 1);
set_show_etc(ingress_rate_cmd,	0xe /*phy*/, 9 /*reg*/, 1);
set_show_etc(ingress_rate_data,	0xe /*phy*/,10 /*reg*/, 1);
set_show_etc(global_status,		0xf /*phy*/, 0 /*reg*/, 1);
set_show_etc(mac01, 			0xf /*phy*/, 1 /*reg*/, 1);
set_show_etc(mac23,				0xf /*phy*/, 2 /*reg*/, 1);
set_show_etc(mac45,				0xf /*phy*/, 3 /*reg*/, 1);
set_show_etc(global_control,	0xf /*phy*/, 4 /*reg*/, 1);
set_show_etc(vtu_control, 		0xf /*phy*/, 5 /*reg*/, 1);
set_show_etc(vtu_vid, 			0xf /*phy*/, 6 /*reg*/, 1);
set_show_etc(vtu_data30, 		0xf /*phy*/, 7 /*reg*/, 1);
set_show_etc(vtu_data54, 		0xf /*phy*/, 8 /*reg*/, 1);
set_show_etc(atu_control, 		0xf /*phy*/,10 /*reg*/, 1);
set_show_etc(atu_operation, 	0xf /*phy*/,11 /*reg*/, 1);
set_show_etc(atu_data, 			0xf /*phy*/,12 /*reg*/, 1);
set_show_etc(atu_mac01, 		0xf /*phy*/,13 /*reg*/, 1);
set_show_etc(atu_mac23, 		0xf /*phy*/,14 /*reg*/, 1);
set_show_etc(atu_mac45, 		0xf /*phy*/,15 /*reg*/, 1);
set_show_etc(free_plus, 		0xf /*phy*/,27 /*reg*/, 1);
set_show_etc(stats_operation, 	0xf /*phy*/,29 /*reg*/, 1);
set_show_etc(stats_data32, 		0xf /*phy*/,30 /*reg*/, 1);
set_show_etc(stats_data10, 		0xf /*phy*/,31 /*reg*/, 1);

static int register_device_files (struct device *dev)
{
	int err = 0;
	
	dev_dbg (dev, "creating device files\n");
	
	// TODO fold device file creates into composite call
	err |= device_create_file(dev, &dev_attr_0_port_status);
	err |= device_create_file(dev, &dev_attr_0_mac_control);
	err |= device_create_file(dev, &dev_attr_0_switch_id);
	err |= device_create_file(dev, &dev_attr_0_port_control);
	err |= device_create_file(dev, &dev_attr_0_port_vlan_map);
	err |= device_create_file(dev, &dev_attr_0_def_vlan_ID_pri);
	err |= device_create_file(dev, &dev_attr_0_port_control2);
	err |= device_create_file(dev, &dev_attr_0_ingr_rate_ctrl);
	err |= device_create_file(dev, &dev_attr_0_egr_rate_ctrl);
	err |= device_create_file(dev, &dev_attr_0_port_assoc_vec);
	err |= device_create_file(dev, &dev_attr_0_rx_frame_cnt);
	err |= device_create_file(dev, &dev_attr_0_tx_frame_cnt);
	err |= device_create_file(dev, &dev_attr_0_rx_pol_cnt);
	err |= device_create_file(dev, &dev_attr_0_queue_cnt);
	err |= device_create_file(dev, &dev_attr_1_port_status);
	err |= device_create_file(dev, &dev_attr_1_mac_control);
	err |= device_create_file(dev, &dev_attr_1_switch_id);
	err |= device_create_file(dev, &dev_attr_1_port_control);
	err |= device_create_file(dev, &dev_attr_1_port_vlan_map);
	err |= device_create_file(dev, &dev_attr_1_def_vlan_ID_pri);
	err |= device_create_file(dev, &dev_attr_1_port_control2);
	err |= device_create_file(dev, &dev_attr_1_ingr_rate_ctrl);
	err |= device_create_file(dev, &dev_attr_1_egr_rate_ctrl);
	err |= device_create_file(dev, &dev_attr_1_port_assoc_vec);
	err |= device_create_file(dev, &dev_attr_1_rx_frame_cnt);
	err |= device_create_file(dev, &dev_attr_1_tx_frame_cnt);
	err |= device_create_file(dev, &dev_attr_1_rx_pol_cnt);
	err |= device_create_file(dev, &dev_attr_1_queue_cnt);
	err |= device_create_file(dev, &dev_attr_2_port_status);
	err |= device_create_file(dev, &dev_attr_2_mac_control);
	err |= device_create_file(dev, &dev_attr_2_switch_id);
	err |= device_create_file(dev, &dev_attr_2_port_control);
	err |= device_create_file(dev, &dev_attr_2_port_vlan_map);
	err |= device_create_file(dev, &dev_attr_2_def_vlan_ID_pri);
	err |= device_create_file(dev, &dev_attr_2_port_control2);
	err |= device_create_file(dev, &dev_attr_2_ingr_rate_ctrl);
	err |= device_create_file(dev, &dev_attr_2_egr_rate_ctrl);
	err |= device_create_file(dev, &dev_attr_2_port_assoc_vec);
	err |= device_create_file(dev, &dev_attr_2_rx_frame_cnt);
	err |= device_create_file(dev, &dev_attr_2_tx_frame_cnt);
	err |= device_create_file(dev, &dev_attr_2_rx_pol_cnt);
	err |= device_create_file(dev, &dev_attr_2_queue_cnt);
	err |= device_create_file(dev, &dev_attr_3_port_status);
	err |= device_create_file(dev, &dev_attr_3_mac_control);
	err |= device_create_file(dev, &dev_attr_3_switch_id);
	err |= device_create_file(dev, &dev_attr_3_port_control);
	err |= device_create_file(dev, &dev_attr_3_port_vlan_map);
	err |= device_create_file(dev, &dev_attr_3_def_vlan_ID_pri);
	err |= device_create_file(dev, &dev_attr_3_port_control2);
	err |= device_create_file(dev, &dev_attr_3_ingr_rate_ctrl);
	err |= device_create_file(dev, &dev_attr_3_egr_rate_ctrl);
	err |= device_create_file(dev, &dev_attr_3_port_assoc_vec);
	err |= device_create_file(dev, &dev_attr_3_rx_frame_cnt);
	err |= device_create_file(dev, &dev_attr_3_tx_frame_cnt);
	err |= device_create_file(dev, &dev_attr_3_rx_pol_cnt);
	err |= device_create_file(dev, &dev_attr_3_queue_cnt);
	err |= device_create_file(dev, &dev_attr_4_port_status);
	err |= device_create_file(dev, &dev_attr_4_mac_control);
	err |= device_create_file(dev, &dev_attr_4_switch_id);
	err |= device_create_file(dev, &dev_attr_4_port_control);
	err |= device_create_file(dev, &dev_attr_4_port_vlan_map);
	err |= device_create_file(dev, &dev_attr_4_def_vlan_ID_pri);
	err |= device_create_file(dev, &dev_attr_4_port_control2);
	err |= device_create_file(dev, &dev_attr_4_ingr_rate_ctrl);
	err |= device_create_file(dev, &dev_attr_4_egr_rate_ctrl);
	err |= device_create_file(dev, &dev_attr_4_port_assoc_vec);
	err |= device_create_file(dev, &dev_attr_4_rx_frame_cnt);
	err |= device_create_file(dev, &dev_attr_4_tx_frame_cnt);
	err |= device_create_file(dev, &dev_attr_4_rx_pol_cnt);
	err |= device_create_file(dev, &dev_attr_4_queue_cnt);
	err |= device_create_file(dev, &dev_attr_5_port_status);
	err |= device_create_file(dev, &dev_attr_5_mac_control);
	err |= device_create_file(dev, &dev_attr_5_switch_id);
	err |= device_create_file(dev, &dev_attr_5_port_control);
	err |= device_create_file(dev, &dev_attr_5_port_vlan_map);
	err |= device_create_file(dev, &dev_attr_5_def_vlan_ID_pri);
	err |= device_create_file(dev, &dev_attr_5_port_control2);
	err |= device_create_file(dev, &dev_attr_5_ingr_rate_ctrl);
	err |= device_create_file(dev, &dev_attr_5_egr_rate_ctrl);
	err |= device_create_file(dev, &dev_attr_5_port_assoc_vec);
	err |= device_create_file(dev, &dev_attr_5_rx_frame_cnt);
	err |= device_create_file(dev, &dev_attr_5_tx_frame_cnt);
	err |= device_create_file(dev, &dev_attr_5_rx_pol_cnt);
	err |= device_create_file(dev, &dev_attr_5_queue_cnt);
	err |= device_create_file(dev, &dev_attr_trunk_mask);
	err |= device_create_file(dev, &dev_attr_ingress_rate_cmd);
	err |= device_create_file(dev, &dev_attr_ingress_rate_data);
	err |= device_create_file(dev, &dev_attr_global_status);
	err |= device_create_file(dev, &dev_attr_mac01);
	err |= device_create_file(dev, &dev_attr_mac23);
	err |= device_create_file(dev, &dev_attr_mac45);
	err |= device_create_file(dev, &dev_attr_global_control);
	err |= device_create_file(dev, &dev_attr_vtu_control);
	err |= device_create_file(dev, &dev_attr_vtu_vid);
	err |= device_create_file(dev, &dev_attr_vtu_data30);
	err |= device_create_file(dev, &dev_attr_vtu_data54);
	err |= device_create_file(dev, &dev_attr_atu_control);
	err |= device_create_file(dev, &dev_attr_atu_operation);
	err |= device_create_file(dev, &dev_attr_atu_data);
	err |= device_create_file(dev, &dev_attr_atu_mac01);
	err |= device_create_file(dev, &dev_attr_atu_mac23);
	err |= device_create_file(dev, &dev_attr_atu_mac45);
	err |= device_create_file(dev, &dev_attr_free_plus);
	err |= device_create_file(dev, &dev_attr_stats_operation);
	err |= device_create_file(dev, &dev_attr_stats_data32);
	err |= device_create_file(dev, &dev_attr_stats_data10);
	
	if (err) {
		dev_err (dev, "device_create_file failed.\n");
	}	
	return err;
}

static int mv88e6061_remove(struct platform_device *pdev)
{
	if (mv88e6061_ptr->base_addr != NULL)
		iounmap(mv88e6061_ptr->base_addr);

	return 0;
}

static int mv88e6061_probe(struct platform_device *pdev)
{
	int err = 0;
	
	if (pdev == NULL)
		return -ENODEV;
	
	mv88e6061_ptr->base_addr = ioremap(pdev->resource[0].start,
			pdev->resource[0].end - pdev->resource[0].start);
	if (mv88e6061_ptr->base_addr == NULL) {
		dev_err(&pdev->dev, "Failed to ioremap ethernet registers\n");
		err = -EIO;
		goto err_out;
	}
	
#if 0	
	/* Let's start by dumping the switch state */
	{
		int i, j;
		u32 regData;
		
		printk("    ");
		for (j = 0; j <=  (1<<5)-1; j++) {
			printk("%04x ", j);
		}
		printk("\n");
		for (i = 0; i <=  (1<<5)-1; i++) {
			printk ("%02x: ", i);
			for (j = 0; j <=  (1<<5)-1; j++) {
 				regData = mdio_read(pdev, j, i);
				printk("%04x ", regData);
			}
			printk("\n");
		}
	}

	{
		typedef struct {
			u8	smiDevLow, smiDevHigh;
			u8	smiReg;
			char* name;
		} regMapEntry;
		regMapEntry regMap[] = {
			{0x0, 0x4, 0x00, "PHY Control     "},
			{0x0, 0x4, 0x01, "PHY Status      "},
			{0x0, 0x4, 0x02, "PHY Identifier  "},
			{0x0, 0x4, 0x03, "PHY Identifier  "},
   			{0x0, 0x4, 0x04, "Auto-Neg Advert "},
			{0x0, 0x4, 0x05, "Link Part Abil  "},
			{0x0, 0x4, 0x06, "Auto-Neg Exp    "},
			{0x0, 0x4, 0x07, "Next Page Trans "},
			{0x0, 0x4, 0x08, "Link Part Nxt Pg"},
			{0x0, 0x4, 0x10, "PHY Spec Ctrl   "},
			{0x0, 0x4, 0x11, "PHY Spec Stat   "},
			{0x0, 0x4, 0x12, "PHY Int En      "},
			{0x0, 0x4, 0x13, "PHY Int Stat    "},
			{0x0, 0x4, 0x14, "Int Port Sum    "},
			{0x0, 0x4, 0x15, "Rx Err Ctr      "},
			{0x0, 0x4, 0x16, "Led Paral Sel   "},
			{0x0, 0x4, 0x17, "LED Stream Sel  "},
			{0x0, 0x4, 0x18, "Led Ctrl        "},

			{0x8, 0xd, 0x00, "Port Status     "},
			{0x8, 0xd, 0x01, "MAC Control     "},
			{0x8, 0xd, 0x03, "Switch Ident    "},
			{0x8, 0xd, 0x04, "Port Control    "},
			{0x8, 0xd, 0x06, "Port-VLAN Map   "},
			{0x8, 0xd, 0x07, "Def VLAN ID/Pri "},
			{0x8, 0xd, 0x08, "Port Control 2  "},
			{0x8, 0xd, 0x09, "Ingr Rate Ctrl  "},
			{0x8, 0xd, 0x0A, "Egr Rate Ctrl   "},
			{0x8, 0xd, 0x0B, "Port Assoc Vec  "},
			{0x8, 0xd, 0x10, "Rx Frm Ctr      "},
			{0x8, 0xd, 0x11, "Tx Frm Ctr      "},
			{0x8, 0xd, 0x12, "Rx Pol Ctr      "},
			{0x8, 0xd, 0x1B, "Queue Ctr       "},

			{0xe, 0xe, 0x07, "Trunk Mask      "},
			{0xe, 0xe, 0x09, "Ingr Rate Ctrl  "},
			{0xe, 0xe, 0x0A, "Ingr Rate Data  "},

			{0xf, 0xf, 0x00, "Global Status   "},
			{0xf, 0xf, 0x01, "Switch MAC 1    "},
			{0xf, 0xf, 0x02, "Switch MAC 2    "},
			{0xf, 0xf, 0x03, "Switch MAC 3    "},
			{0xf, 0xf, 0x04, "Global Control  "},
			{0xf, 0xf, 0x05, "VTU Control     "},
			{0xf, 0xf, 0x06, "VTU VID         "},
			{0xf, 0xf, 0x07, "VTU Data Prt 3:0"},
			{0xf, 0xf, 0x08, "VTU Data Prt 5:4"},
			{0xf, 0xf, 0x0A, "ATU Control     "},
			{0xf, 0xf, 0x0B, "ATU Operation   "},
			{0xf, 0xf, 0x0C, "ATU Data        "},
			{0xf, 0xf, 0x0D, "ATU MAC 1       "},
			{0xf, 0xf, 0x0E, "ATU MAC 2       "},
			{0xf, 0xf, 0x0F, "ATU MAC 3       "},
			{0xf, 0xf, 0x10, "IP PRI 1        "},
			{0xf, 0xf, 0x11, "IP PRI 2        "},
			{0xf, 0xf, 0x12, "IP PRI 3        "},
			{0xf, 0xf, 0x13, "IP PRI 4        "},
			{0xf, 0xf, 0x14, "IP PRI 5        "},
			{0xf, 0xf, 0x15, "IP PRI 6        "},
			{0xf, 0xf, 0x16, "IP PRI 7        "},
			{0xf, 0xf, 0x17, "IP PRI 8        "},
			{0xf, 0xf, 0x1B, "Free Plus       "},
			{0xf, 0xf, 0x1D, "Stats Operation "},
			{0xf, 0xf, 0x1E, "Stats Data 3:2  "},
			{0xf, 0xf, 0x1F, "Stats Data 1:0  "},
   		};
		
		int i, j;
		u32 regData;
		
		for (i = 0; i < sizeof(regMap)/sizeof(regMap[0]); i++) {
			printk ("%02x %s [%02x-%02x]:\t", regMap[i].smiReg, regMap[i].name, regMap[i].smiDevLow, regMap[i].smiDevHigh);
			for (j = regMap[i].smiDevLow; j <= regMap[i].smiDevHigh; j++) {
 				regData = mdio_read(pdev, j, regMap[i].smiReg);
				printk ("%04x ", regData);
				udelay(100);
			}
			printk("\n");
		}
	}
#endif
	
	register_device_files (&pdev->dev);
	
	return 0;

err_out:
	mv88e6061_remove(pdev);
	return err;
}

static struct platform_driver mv88e6061_driver = {
	.probe		= mv88e6061_probe,
	.remove		= mv88e6061_remove,
	.driver		= {
		.name	= "mv88e6061-switch",
	},
};

static int __init mv88e6061_init_module(void)
{
	printk(KERN_INFO DRV_MODULE_NAME " version " DRV_MODULE_VERSION " loading\n");
	return platform_driver_register(&mv88e6061_driver);
}

static void __exit mv88e6061_cleanup_module(void)
{
	platform_driver_unregister(&mv88e6061_driver);
}

module_init(mv88e6061_init_module);
module_exit(mv88e6061_cleanup_module);

MODULE_DESCRIPTION("Marvell 88E6061 Switch driver");
MODULE_AUTHOR("Theuns Verwoerd <theuns@bluewatersys.com>");
MODULE_LICENSE("GPL");
