/*======================================================================

  Device driver for the PCMCIA control functionality of EP93xx
  microprocessors.

    The contents of this file may be used under the
    terms of the GNU Public License version 2 (the "GPL")
  
    (c) Ian Molton (spyro@f2s.com) 2003
    (c) Stefan Eletzhofer (stefan.eletzhofer@inquant.de) 2003,4
    (c) Robert Whaley (rwhaley@applieddata.net) 2005

    derived from sa11xx_base.c

     Portions created by John G. Dorsey are
     Copyright (C) 1999 John G. Dorsey.

    Updated for mainline kernel:
    (C) 2009 Ryan Mallon <ryan@bluewatersys.com> 
         
  ======================================================================*/

#include <linux/init.h>
#include <linux/platform_device.h>

#include <pcmcia/cs_types.h>
#include <pcmcia/ss.h>
#include <pcmcia/cistpl.h>

#include <mach/hardware.h>
#include <asm/io.h>

#include "ep93xx_pcmcia.h"
#include "cs_internal.h"
#include "soc_common.h"

#define MCXX_SETUP_MAX     		((u_int)0xff)
#define MCXX_ASST_MAX      		((u_int)0xff)
#define MCXX_HOLD_MAX      		((u_int)0x0f)

#define MCXX_SETUP_SHIFT   		0
#define MCXX_ASST_SHIFT    		16
#define MCXX_HOLD_SHIFT    		8

struct ep93xx_pcmcia {
	ep93xx_pcmcia_probe_func_t	probe_func;
	struct pcmcia_low_level		*ops;
	void __iomem	 		*regs;
	struct pcmcia_irqs		*irqs;
	int				nr_irqs;
	int	 			*socket_irq;
};

static struct ep93xx_pcmcia ep93xx_pcmcia;

void ep93xx_pcmcia_write(unsigned int val, unsigned int reg)
{
	__raw_writel(val, ep93xx_pcmcia.regs + reg);
}

unsigned int ep93xx_pcmcia_read(unsigned int reg)
{
	return __raw_readl(ep93xx_pcmcia.regs + reg);
}

static inline u_int ep93xx_mcxx_hold(u_int pcmcia_cycle_ns, u_int hclk_10khz)
{
	u_int code = pcmcia_cycle_ns * hclk_10khz;
	return (code / 100000) + ((code % 100000) ? 1 : 0) - 1;
}

static inline u_int ep93xx_mcxx_asst(u_int pcmcia_cycle_ns, u_int hclk_10khz)
{
	u_int code = pcmcia_cycle_ns * hclk_10khz;
	return (code / 100000) + ((code % 100000) ? 1 : 0) - 1;
}

static inline u_int ep93xx_mcxx_setup(u_int pcmcia_cycle_ns, u_int hclk_10khz)
{
	u_int code = pcmcia_cycle_ns * hclk_10khz;
	return (code / 100000) + ((code % 100000) ? 1 : 0) - 1;
}

static int ep93xx_pcmcia_set_mcmem(int sock, int speed, int clock)
{
        ep93xx_pcmcia_write(EP93XX_PCMCIA_16BITS
			    | (min(ep93xx_mcxx_setup(speed, clock), 
				   MCXX_SETUP_MAX) << MCXX_SETUP_SHIFT)
			    | (min(ep93xx_mcxx_asst(speed, clock), 
				   MCXX_ASST_MAX) << MCXX_ASST_SHIFT)
			    | (min(ep93xx_mcxx_hold(speed, clock), 
				   MCXX_HOLD_MAX) << MCXX_HOLD_SHIFT),
			    EP93XX_PCCOMMON);
	
	return 0;
}

static int ep93xx_pcmcia_set_mcio(int sock, int speed, int clock)
{
        ep93xx_pcmcia_write(EP93XX_PCMCIA_16BITS
			    | (min(ep93xx_mcxx_setup(speed, clock), 
				   MCXX_SETUP_MAX) << MCXX_SETUP_SHIFT)
			    | (min(ep93xx_mcxx_asst(speed, clock), 
				   MCXX_ASST_MAX) << MCXX_ASST_SHIFT)
			    | (min(ep93xx_mcxx_hold(speed, clock), 
				   MCXX_HOLD_MAX) << MCXX_HOLD_SHIFT),
			    EP93XX_PCIO);
	
	return 0;
}

static int ep93xx_pcmcia_set_mcatt(int sock, int speed, int clock)
{
        ep93xx_pcmcia_write(EP93XX_PCMCIA_16BITS
			    | (min(ep93xx_mcxx_setup(speed, clock), 
				   MCXX_SETUP_MAX) << MCXX_SETUP_SHIFT)
			    | (min(ep93xx_mcxx_asst(speed, clock), 
				   MCXX_ASST_MAX) << MCXX_ASST_SHIFT)
			    | (min(ep93xx_mcxx_hold(speed, clock), 
				   MCXX_HOLD_MAX) << MCXX_HOLD_SHIFT),
			    EP93XX_PCATTRIB);
	
	return 0;
}

static int ep93xx_pcmcia_set_mcxx(struct soc_pcmcia_socket *skt, 
				  unsigned int clk)
{
	struct soc_pcmcia_timing timing;
	int sock = skt->nr;
	
	soc_common_pcmcia_get_timing(skt, &timing);
	
	ep93xx_pcmcia_set_mcmem(sock, timing.mem, clk);
	ep93xx_pcmcia_set_mcatt(sock, timing.attr, clk);
	ep93xx_pcmcia_set_mcio(sock, timing.io, clk);
	
	return 0;
}

static int ep93xx_pcmcia_set_timing(struct soc_pcmcia_socket *skt)
{
	/* FIXME: for now just hard code hclk to 100Mhz (in 10khz units) */
	unsigned int hclk = 10000; 
	return ep93xx_pcmcia_set_mcxx(skt, hclk);
}

void ep93xx_pcmcia_set_irqs(int *socket_irq, struct pcmcia_irqs *irqs, 
			    int nr_irqs)
{
	ep93xx_pcmcia.socket_irq = socket_irq;
	ep93xx_pcmcia.irqs = irqs;
	ep93xx_pcmcia.nr_irqs = nr_irqs;
}

int ep93xx_pcmcia_hw_init(struct soc_pcmcia_socket *skt)
{
	skt->irq = ep93xx_pcmcia.socket_irq[skt->nr];
	if (ep93xx_pcmcia.irqs) 
		return soc_pcmcia_request_irqs(skt, ep93xx_pcmcia.irqs, 
					       ep93xx_pcmcia.nr_irqs);
	else
		return 0;
}

void ep93xx_pcmcia_hw_shutdown(struct soc_pcmcia_socket *skt)
{
	if (ep93xx_pcmcia.irqs)
		soc_pcmcia_free_irqs(skt, ep93xx_pcmcia.irqs, 
				     ep93xx_pcmcia.nr_irqs);
}

void ep93xx_pcmcia_socket_state(struct soc_pcmcia_socket *skt,
				struct pcmcia_state *state)
{
	unsigned long status = __raw_readl(EP93XX_GPIO_PFDR);

	state->detect	= status & EP93XX_PCMCIA_DETECT1	? 0 : 1;
	state->ready	= status & EP93XX_PCMCIA_READY		? 1 : 0;
	state->wrprot	= status & EP93XX_PCMCIA_WRPROT		? 0 : 1;
	state->bvd1	= status & EP93XX_PCMCIA_BVD1		? 1 : 0;
	state->bvd2	= status & EP93XX_PCMCIA_BVD2		? 1 : 0;
	state->vs_3v	= status & EP93XX_PCMCIA_VS_3V		? 0 : 1;
	state->vs_Xv	= status & EP93XX_PCMCIA_VS_XV		? 0 : 1;
}

int ep93xx_pcmcia_configure_socket(struct soc_pcmcia_socket *skt,
				   const socket_state_t *state)
{
	unsigned int reg;

	reg = ep93xx_pcmcia_read(EP93XX_PCCTRL);
	if (state->flags & SS_RESET)
		reg |= EP93XX_PCCTRL_RESET;
	else
		reg &= ~EP93XX_PCCTRL_RESET;
	ep93xx_pcmcia_write(reg, EP93XX_PCCTRL);
	return 0;
}

void ep93xx_pcmcia_socket_init(struct soc_pcmcia_socket *skt)
{
	/* Use compact flash mode */
	ep93xx_pcmcia_write(EP93XX_PCCTRL_MODE_CF | 
			    EP93XX_PCCTRL_WAIT_ENABLE, EP93XX_PCCTRL);
	soc_pcmcia_enable_irqs(skt, ep93xx_pcmcia.irqs, ep93xx_pcmcia.nr_irqs);
}

void ep93xx_pcmcia_socket_suspend(struct soc_pcmcia_socket *skt)
{
	soc_pcmcia_disable_irqs(skt, ep93xx_pcmcia.irqs, 
				ep93xx_pcmcia.nr_irqs);
}

static int ep93xx_pcmcia_suspend(struct platform_device *pdev, 
				 pm_message_t state)
{
	return pcmcia_socket_dev_suspend(&pdev->dev, state);
}

static int ep93xx_pcmcia_resume(struct platform_device *pdev)
{
	return pcmcia_socket_dev_resume(&pdev->dev);
}

static int __ep93xx_pcmcia_probe(struct platform_device *pdev)
{
	struct pcmcia_low_level *ops = ep93xx_pcmcia.ops;
	struct resource *res;
	int err;

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res)
		return -ENXIO;

	res = request_mem_region(res->start, resource_size(res), pdev->name);
	if (!res)
		return -EBUSY;
	
	ep93xx_pcmcia.regs = ioremap(res->start, resource_size(res));
	if (!ep93xx_pcmcia.regs) {
		err = -ENXIO;
		goto fail_ioremap;
	}
	
       	/* Provide our EP93XX specific timing routines */
	ops->set_timing = ep93xx_pcmcia_set_timing;
	err = soc_common_drv_pcmcia_probe(&pdev->dev, ops, 
					  ops->first, ops->nr);
	if (err)
		goto fail_soc_probe;
		
	if (ep93xx_pcmcia.probe_func) {
		err = ep93xx_pcmcia.probe_func(&pdev->dev);
		if (err)
			goto fail_board_init;
	}
	
	return 0;

fail_board_init:
	/* FIXME - unprobe? */
fail_soc_probe:	
	iounmap(ep93xx_pcmcia.regs);	
fail_ioremap:
	release_mem_region(res->start, resource_size(res));
	return err;
}

static struct platform_driver ep93xx_pcmcia_driver = {
	.driver		= {
		.name	= "ep93xx-pcmcia",
		.owner	= THIS_MODULE,
	},
	.remove		= __exit_p(ep93xx_pcmcia_remove),
	.suspend	= ep93xx_pcmcia_suspend,
	.resume		= ep93xx_pcmcia_resume,
};

int ep93xx_pcmcia_probe(ep93xx_pcmcia_probe_func_t probe_func,
			struct pcmcia_low_level *ops)
{	
	ep93xx_pcmcia.ops = ops;
	ep93xx_pcmcia.probe_func = probe_func;
	return platform_driver_probe(&ep93xx_pcmcia_driver,
				     __ep93xx_pcmcia_probe);
}

void ep93xx_pcmcia_unregister(void)
{
	platform_driver_unregister(&ep93xx_pcmcia_driver);
}

