/*
 * drivers/pcmcia/ep93xx_rig200.c
 *
 * Compact flash support on Snapper CL15/Rig200
 * 
 * Author:	Ryan Mallon <ryan@bluewatersys.com>
 * Copyright:	2009 Bluewater Systems
 *
 * The Rig200 has two compact flash 'slots', a compact flash socket, and a
 * compact flash wifi module. On the EP93xx, only a single compact flash 
 * socket is supported. The active socket can be selected using the cf_slot
 * sysfs file.
 *   slot 0: Compact flash socket
 *   slot 1: Wifi module
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/gpio.h>

#include <asm/hardware/snapper.h>
#include <mach/hardware.h>
#include <mach/irqs.h>

#include "ep93xx_pcmcia.h"

static struct pcmcia_irqs irqs[] = {
	{0, gpio_to_irq(EP93XX_GPIO_LINE_F(1)), "CF Card Detect"},
	{0, gpio_to_irq(EP93XX_GPIO_LINE_F(4)), "CF BVD"},
};

static int socket_irq[] = {gpio_to_irq(EP93XX_GPIO_LINE_F(6))};

static inline int rig200_pcmcia_set_slot(int slot)
{
	struct soc_pcmcia_socket *skt;
	
	list_for_each_entry(skt, &soc_pcmcia_sockets, node)
		if (skt->socket.pcmcia_state.present) {
			dev_err(skt->dev, 
				"Cannot change slot while card is present\n");
			return -EBUSY;
		}

	gpio_set_value(EP93XX_GPIO_LINE_B(6), slot);	
	return 0;
}

static inline int rig200_pcmcia_get_slot(void)
{
	return gpio_get_value(EP93XX_GPIO_LINE_B(6));
}

static ssize_t rig200_sysfs_cf_slot_show(struct device *dev, 
					 struct device_attribute *attr, 
					 char *buf)
{
	return sprintf(buf, "%d\n", rig200_pcmcia_get_slot());
}

static ssize_t rig200_sysfs_cf_slot_store(struct device *dev, 
					  struct device_attribute *attr,
					  const char *buf, size_t count)
{
	int err, slot = !!simple_strtol(buf, NULL, 0);
	
	err = rig200_pcmcia_set_slot(slot);
	return err ? err : count;
}

static DEVICE_ATTR(cf_slot, S_IWUSR | S_IRUGO, rig200_sysfs_cf_slot_show,
		   rig200_sysfs_cf_slot_store);

static void rig200_pcmcia_socket_state(struct soc_pcmcia_socket *skt,
				       struct pcmcia_state *state)
{
	unsigned long status;

	status = __raw_readl(EP93XX_GPIO_PFDR);
	switch (rig200_pcmcia_get_slot()) {
	case 0:
		ep93xx_pcmcia_socket_state(skt, state);
		break;
	case 1:		
		state->detect	= status & EP93XX_PCMCIA_DETECT2 ? 0 : 1;
		state->ready	= status & EP93XX_PCMCIA_READY	 ? 1 : 0;
		state->wrprot	= status & EP93XX_PCMCIA_WRPROT	 ? 0 : 1;
		state->vs_3v	= 1;
		state->bvd1	= 0;
		state->bvd2	= 0;
		break;
	}	
}

static struct pcmcia_low_level rig200_pcmcia_ops = {
	.owner			= THIS_MODULE,
	.hw_init		= ep93xx_pcmcia_hw_init,
	.hw_shutdown		= ep93xx_pcmcia_hw_shutdown,
	.socket_state		= rig200_pcmcia_socket_state,
	.configure_socket	= ep93xx_pcmcia_configure_socket,
	.socket_init		= ep93xx_pcmcia_socket_init,
	.socket_suspend		= ep93xx_pcmcia_socket_suspend,
	.first			= 0,
	.nr			= 1,
};

static int rig200_pcmcia_probe(struct device *dev)
{
	int err;

	err = gpio_request(EP93XX_GPIO_LINE_B(6), "cf_slot");
	if (err)
		goto out_no_gpio;
	
	/* Default to compact flash slot */
	gpio_direction_output(EP93XX_GPIO_LINE_B(6), 0);

	err = device_create_file(dev, &dev_attr_cf_slot);
	if (err)
		goto out_no_sysfs;
	
	return 0;

out_no_sysfs:
	gpio_free(EP93XX_GPIO_LINE_B(6));
out_no_gpio:
	/* FIXME - Remove device */
	return err;
}

static int __init rig200_pcmcia_init(void)
{
	if (!snapper_baseboard_is_rig200()) 
		return -ENODEV;

	ep93xx_pcmcia_set_irqs(socket_irq, irqs, ARRAY_SIZE(irqs));
	return ep93xx_pcmcia_probe(rig200_pcmcia_probe, &rig200_pcmcia_ops);
}

static void __exit rig200_pcmcia_exit(void)
{
	ep93xx_pcmcia_unregister();
}

module_init(rig200_pcmcia_init);
module_exit(rig200_pcmcia_exit);

MODULE_AUTHOR("Ryan Mallon <ryan@bluewatersys.com>");
MODULE_DESCRIPTION("Snapper CL15/Rig 200 compact flash driver");
MODULE_LICENSE("GPL");

