/*
 * linux/sound/soc/ep93xx-i2s.c
 * EP93xx I2S driver
 *
 * Copyright (C) 2008 Ryan Mallon <ryan@bluewatersys.com>
 * 
 * Based on the original driver by:
 * Copyright (C) 2007 Chase Douglas <chasedouglas at gmail>
 * Copyright (C) 2006 Lennert Buytenhek <buytenh at wantstofly.org>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/wait.h>
#include <linux/delay.h>
#include <linux/mutex.h>
#include <linux/clk.h>

#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/initval.h>
#include <sound/soc.h>

#include <asm/io.h>

#include <mach/ep93xx-regs.h>
#include <mach/platform.h>
#include <mach/dma.h>
#include <mach/gpio.h>

#include "ep93xx-pcm.h"
#include "ep93xx-i2s.h"

static struct ep93xx_pcm_dma_params ep93xx_i2s_pcm_out = {
	.name		= "i2s-pcm-out",
	.dma_port	= EP93XX_DMA_M2P_PORT_I2S1,
};

static struct ep93xx_pcm_dma_params ep93xx_i2s_pcm_in = {
	.name		= "i2s-pcm-in",
	.dma_port	= EP93XX_DMA_M2P_PORT_I2S1,
};

struct ep93xx_i2s_info {
	struct clk			*clk;
	struct ep93xx_pcm_dma_params	*dma_params[2];	
};

static struct ep93xx_i2s_info ep93xx_i2s;

static inline void ep93xx_i2s_writel(u32 val, u32 reg)
{
	__raw_writel(val, reg);
}

static inline u32 ep93xx_i2s_readl(u32 reg)
{
	return __raw_readl(reg);
}

static inline void ep93xx_i2s_enable(void)
{
	int i;

	/* Drain the FIFOs */
	for (i = 0; i < 8; i++) {
		ep93xx_i2s_writel(0x0, EP93XX_I2S_TX0LFT);
		ep93xx_i2s_writel(0x0, EP93XX_I2S_TX0RT);
		ep93xx_i2s_writel(0x0, EP93XX_I2S_TX1LFT);
		ep93xx_i2s_writel(0x0, EP93XX_I2S_TX1RT);
		ep93xx_i2s_writel(0x0, EP93XX_I2S_TX2LFT);
		ep93xx_i2s_writel(0x0, EP93XX_I2S_TX2RT);
	}
	
	/* Enable the FIFOs */
	ep93xx_i2s_writel(0x1, EP93XX_I2S_RX0EN);
	ep93xx_i2s_writel(0x1, EP93XX_I2S_RX1EN);
	ep93xx_i2s_writel(0x1, EP93XX_I2S_RX2EN);
	ep93xx_i2s_writel(0x1, EP93XX_I2S_TX0EN);
	ep93xx_i2s_writel(0x1, EP93XX_I2S_TX1EN);
	ep93xx_i2s_writel(0x1, EP93XX_I2S_TX2EN);

	ep93xx_i2s_writel(0x1, EP93XX_I2S_GLCTRL);
}

static inline void ep93xx_i2s_disable(void)
{
	/* Disable the FIFOs */
	ep93xx_i2s_writel(0x0, EP93XX_I2S_RX0EN);
	ep93xx_i2s_writel(0x0, EP93XX_I2S_RX1EN);
	ep93xx_i2s_writel(0x0, EP93XX_I2S_RX2EN);
	ep93xx_i2s_writel(0x0, EP93XX_I2S_TX0EN);
	ep93xx_i2s_writel(0x0, EP93XX_I2S_TX1EN);
	ep93xx_i2s_writel(0x0, EP93XX_I2S_TX2EN);

	ep93xx_i2s_writel(0x0, EP93XX_I2S_GLCTRL);
}

static int ep93xx_i2s_set_dai_sysclk(struct snd_soc_dai *cpu_dai, int clk_id,
				     unsigned int freq, int dir)
{
	u32 val;
	
	clk_set_rate(ep93xx_i2s.clk, freq);	

	/* 
	 * Enable i2s slave mode
	 * FIXME - Should be done in set_dai_fmt
	 */
	val = __raw_readl(EP93XX_SYSCON_I2SCLKDIV);
	ep93xx_syscon_swlocked_write(EP93XX_SYSCON_I2SCLKDIV, 
				     val | 0xa00b0000);
	return 0;
}

static int ep93xx_i2s_startup(struct snd_pcm_substream *substream,
			      struct snd_soc_dai *dai)
{	
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_dai *cpu_dai = rtd->dai->cpu_dai;

	cpu_dai->dma_data = ep93xx_i2s.dma_params[substream->stream];
	return 0;
}

static int ep93xx_i2s_trigger(struct snd_pcm_substream *substream, int cmd,
			      struct snd_soc_dai *dai)
{
	int ret = 0;

	switch (cmd) {
	case SNDRV_PCM_TRIGGER_START:
		break;
	case SNDRV_PCM_TRIGGER_RESUME:
	case SNDRV_PCM_TRIGGER_PAUSE_RELEASE:
	case SNDRV_PCM_TRIGGER_STOP:
	case SNDRV_PCM_TRIGGER_SUSPEND:
	case SNDRV_PCM_TRIGGER_PAUSE_PUSH:
		break;
	default:
		ret = -EINVAL;
	}
	
	return ret;
}

static void ep93xx_i2s_shutdown(struct snd_pcm_substream *substream,
				struct snd_soc_dai *dai)
{	
	clk_disable(ep93xx_i2s.clk);
	ep93xx_i2s_disable();
}

static int ep93xx_i2s_set_dai_fmt(struct snd_soc_dai *cpu_dai, 
				  unsigned int fmt)
{
	unsigned int val;
	
	switch (fmt & SND_SOC_DAIFMT_FORMAT_MASK) {
	case SND_SOC_DAIFMT_I2S:
		val = ep93xx_i2s_readl(EP93XX_I2S_RXCLKCFG);
		val |= EP93XX_I2S_CLKCFG_TREL | EP93XX_I2S_CLKCFG_NBCG;
		ep93xx_i2s_writel(val, EP93XX_I2S_RXCLKCFG);
		ep93xx_i2s_writel(val, EP93XX_I2S_TXCLKCFG);	
		break;
		
	case SND_SOC_DAIFMT_LEFT_J:
		val = ep93xx_i2s_readl(EP93XX_I2S_RXLINCTRLDATA);
		val &= ~EP93XX_I2S_LINCTRLDATA_R_JUST;
		ep93xx_i2s_writel(val, EP93XX_I2S_RXLINCTRLDATA);
		ep93xx_i2s_writel(val, EP93XX_I2S_TXLINCTRLDATA);	
		break;
		
	case SND_SOC_DAIFMT_RIGHT_J:
		val = ep93xx_i2s_readl(EP93XX_I2S_RXLINCTRLDATA);
		val |= EP93XX_I2S_LINCTRLDATA_R_JUST;
		ep93xx_i2s_writel(val, EP93XX_I2S_RXLINCTRLDATA);
		ep93xx_i2s_writel(val, EP93XX_I2S_TXLINCTRLDATA);
		break;

	default:
		return -EINVAL;
	}

	switch (fmt & SND_SOC_DAIFMT_MASTER_MASK) {
	case SND_SOC_DAIFMT_CBS_CFS:
		val = ep93xx_i2s_readl(EP93XX_I2S_RXCLKCFG);
		val |= EP93XX_I2S_CLKCFG_MASTER;
		ep93xx_i2s_writel(val, EP93XX_I2S_RXCLKCFG);
		ep93xx_i2s_writel(val, EP93XX_I2S_TXCLKCFG);
		break;

	case SND_SOC_DAIFMT_CBM_CFM:
		val = ep93xx_i2s_readl(EP93XX_I2S_RXCLKCFG);
		val &= ~EP93XX_I2S_CLKCFG_MASTER;
		ep93xx_i2s_writel(val, EP93XX_I2S_RXCLKCFG);
		ep93xx_i2s_writel(val, EP93XX_I2S_TXCLKCFG);
		break;

	default:
		return -EINVAL;
	}

	return 0;
}

static int ep93xx_i2s_hw_params(struct snd_pcm_substream *substream,
				struct snd_pcm_hw_params *params,
				struct snd_soc_dai *dai)
{
	switch (params_format(params)) {
	case SNDRV_PCM_FORMAT_S16_LE:
		ep93xx_i2s_writel(EP93XX_I2S_WRDLEN_16, EP93XX_I2S_TXWRDLEN);
		ep93xx_i2s_writel(EP93XX_I2S_WRDLEN_16, EP93XX_I2S_TXWRDLEN);
		break;

	case SNDRV_PCM_FORMAT_S24_LE:
		ep93xx_i2s_writel(EP93XX_I2S_WRDLEN_24, EP93XX_I2S_TXWRDLEN);
		ep93xx_i2s_writel(EP93XX_I2S_WRDLEN_24, EP93XX_I2S_TXWRDLEN);
		break;

	case SNDRV_PCM_FORMAT_S32_LE:
		ep93xx_i2s_writel(EP93XX_I2S_WRDLEN_32, EP93XX_I2S_TXWRDLEN);
		ep93xx_i2s_writel(EP93XX_I2S_WRDLEN_32, EP93XX_I2S_TXWRDLEN);
		break;

	default:
		return -EINVAL;
	}

	clk_enable(ep93xx_i2s.clk);
	ep93xx_i2s_enable();
	return 0;
}

#ifdef CONFIG_PM
static int ep93xx_i2s_suspend(struct snd_soc_dai *dai)
{
	if (!dai->active)
		return;

	ep93xx_i2s_enable();
	clk_disable(ep93xx_i2s.clk);
}

static int ep93xx_i2s_resume(struct snd_soc_dai *dai)
{
	if (!dai->active)
		return;

	ep93xx_i2s_disable();
	clk_enable(ep93xx_i2s.clk);
}
#else
#define ep93xx_i2s_suspend	NULL
#define ep93xx_i2s_resume	NULL
#endif

#define EP93XX_I2S_FORMATS (SNDRV_PCM_FMTBIT_S16_LE | \
			    SNDRV_PCM_FMTBIT_S24_LE | \
			    SNDRV_PCM_FMTBIT_S32_LE)

struct snd_soc_dai ep93xx_i2s_dai = {
	.name		= "ep93xx-i2s",
	.id		= 0,
	.suspend	= ep93xx_i2s_suspend,
	.resume		= ep93xx_i2s_resume,
	.playback	= {
		.channels_min	= 2,
		.channels_max	= 2,
		.rates		= SNDRV_PCM_RATE_8000_48000,
		.formats	= EP93XX_I2S_FORMATS,
	},
	.capture	= {
		 .channels_min	= 2,
		 .channels_max	= 2,
		 .rates		= SNDRV_PCM_RATE_8000_48000,
		 .formats	= EP93XX_I2S_FORMATS,
	},
	.ops		= {
		 .startup	= ep93xx_i2s_startup,
		 .shutdown	= ep93xx_i2s_shutdown,
		 .trigger	= ep93xx_i2s_trigger,
		 .hw_params	= ep93xx_i2s_hw_params,
		 .set_fmt	= ep93xx_i2s_set_dai_fmt,
		 .set_sysclk	= ep93xx_i2s_set_dai_sysclk,
	},
};
EXPORT_SYMBOL_GPL(ep93xx_i2s_dai);

static int ep93xx_i2s_probe(struct platform_device *pdev)
{	
	int err;

	ep93xx_i2s.clk = clk_get(&pdev->dev, "i2s");
	if (IS_ERR(ep93xx_i2s.clk)) {
		err = PTR_ERR(ep93xx_i2s.clk);
		goto fail_clk;
	}
	
	ep93xx_i2s.dma_params[SNDRV_PCM_STREAM_PLAYBACK] = &ep93xx_i2s_pcm_out;
	ep93xx_i2s.dma_params[SNDRV_PCM_STREAM_CAPTURE] = &ep93xx_i2s_pcm_in;
	
	ep93xx_i2s_dai.dev = &pdev->dev;
	err = snd_soc_register_dai(&ep93xx_i2s_dai);
	if (err)
		goto fail_register;

	return 0;
	
fail_register:
	clk_put(ep93xx_i2s.clk);
fail_clk:
	return err;
}

static int __devexit ep93xx_i2s_remove(struct platform_device *pdev)
{
	clk_put(ep93xx_i2s.clk);
	snd_soc_unregister_dai(&ep93xx_i2s_dai);
	return 0;
}

static struct platform_driver ep93xx_i2s_driver = {
	.probe		= ep93xx_i2s_probe,
	.remove		= __devexit_p(ep93xx_i2s_remove),
	.driver		= {
		.name	= "ep93xx-i2s",
		.owner	= THIS_MODULE,
	},
};

static int __init ep93xx_i2s_init(void)
{
	return platform_driver_register(&ep93xx_i2s_driver);
}

static void __exit ep93xx_i2s_exit(void)
{
	platform_driver_unregister(&ep93xx_i2s_driver);
}

module_init(ep93xx_i2s_init);
module_exit(ep93xx_i2s_exit);

MODULE_AUTHOR("Ryan Mallon <ryan@bluewatersys.com>");
MODULE_DESCRIPTION("EP93XX I2S driver");
MODULE_LICENSE("GPL");
